# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time
from collections import defaultdict
from typing import List, DefaultDict, Dict, Tuple

from .... import oscar as mo
from ....typing import BandType

logger = logging.getLogger(__name__)


class GlobalSlotManagerActor(mo.Actor):
    # {(address, resource_type): {(session_id, subtask_id): slot_id}}
    _band_stid_slots: DefaultDict[BandType, Dict[Tuple[str, str], int]]
    _band_used_slots: DefaultDict[BandType, int]
    _band_total_slots: Dict[BandType, int]

    def __init__(self):
        self._band_stid_slots = defaultdict(dict)
        self._band_used_slots = defaultdict(lambda: 0)
        self._band_idle_start_time = dict()
        self._band_idle_events = dict()
        self._band_total_slots = dict()

        self._cluster_api = None

        self._band_watch_task = None

    async def __post_create__(self):
        from ...cluster.api import ClusterAPI

        self._cluster_api = await ClusterAPI.create(self.address)

        async def watch_bands():
            async for bands in self._cluster_api.watch_all_bands():
                old_bands = set(self._band_total_slots.keys())
                self._band_total_slots = bands
                new_bands = set(bands.keys()) - old_bands
                for band in new_bands:
                    self._update_slot_usage(band, 0)

        self._band_watch_task = asyncio.create_task(watch_bands())

    async def __pre_destroy__(self):
        self._band_watch_task.cancel()

    async def refresh_bands(self):
        self._band_total_slots = await self._cluster_api.get_all_bands()

    @mo.extensible
    async def apply_subtask_slots(
        self,
        band: BandType,
        session_id: str,
        subtask_ids: List[str],
        subtask_slots: List[int],
    ) -> List[str]:
        if not self._band_total_slots or band not in self._band_total_slots:
            self._band_total_slots = await self._cluster_api.get_all_bands()

        idx = 0
        # only ready bands will pass
        if band in self._band_total_slots:
            total_slots = self._band_total_slots[band]
            for stid, slots in zip(subtask_ids, subtask_slots):
                if self._band_used_slots[band] + slots > total_slots:
                    break
                self._band_stid_slots[band][(session_id, stid)] = slots
                self._update_slot_usage(band, slots)
                idx += 1
        if idx == 0:
            logger.debug(
                "No slots available, status: %r, request: %r",
                self._band_used_slots,
                subtask_slots,
            )
        return subtask_ids[:idx]

    @mo.extensible
    def update_subtask_slots(
        self, band: BandType, session_id: str, subtask_id: str, slots: int
    ):
        session_subtask_id = (session_id, subtask_id)
        subtask_slots = self._band_stid_slots[band]

        if session_subtask_id not in subtask_slots:
            return

        slots_delta = slots - subtask_slots[session_subtask_id]
        subtask_slots[session_subtask_id] = slots
        self._update_slot_usage(band, slots_delta)

    @mo.extensible
    def release_subtask_slots(self, band: BandType, session_id: str, subtask_id: str):
        # todo ensure slots released when subtasks ends in all means
        slots_delta = self._band_stid_slots[band].pop((session_id, subtask_id), 0)
        self._update_slot_usage(band, -slots_delta)

    def _update_slot_usage(self, band: BandType, slots_usage_delta: float):
        self._band_used_slots[band] += slots_usage_delta
        if self._band_used_slots[band] == 0:
            self._band_used_slots.pop(band)
            self._band_idle_start_time[band] = time.time()
            if band in self._band_idle_events:
                self._band_idle_events.pop(band).set()
        else:
            self._band_idle_start_time[band] = -1

    def get_used_slots(self) -> Dict[BandType, int]:
        return self._band_used_slots

    async def get_idle_bands(self, idle_duration: int):
        """Return a band list which all bands has been idle for at least `idle_duration` seconds."""
        now = time.time()
        idle_bands = []
        for band in self._band_total_slots.keys():
            idle_start_time = self._band_idle_start_time[band]
            if idle_start_time > 0 and now >= idle_start_time + idle_duration:
                idle_bands.append(band)
        return idle_bands

    async def wait_band_idle(self, band: BandType):
        if self._band_idle_start_time[band] <= 0:
            if band in self._band_idle_events:
                event = self._band_idle_events[band]
            else:
                event = asyncio.Event()
                self._band_idle_events[band] = event
            return event.wait()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import itertools
from numbers import Integral
from typing import Optional
import uuid

import numpy as np

from ...core import tile
from ...tensor.indexing.core import process_index, calc_shape
from ...tensor.indexing.getitem import TensorIndex


def indexing_to_chunk_indices(output_chunk):
    """
    Compute input_indices and value_indices when read from or write to
    a tensor chunk.

    Parameters
    ----------
    output_chunk:
        A chunk in the output of the `__setitem__` op.

    Returns
    -------
        The indices in the input chunk, and value_indices in the value block
        that will be assigned.
    """
    input_indices = []  # index in the chunk of the mutable tensor
    value_indices = []  # index in the chunk of the assigned value
    for d, s in zip(output_chunk.op.indexes, output_chunk.op.inputs[0].shape):
        # expand the index (slice)
        idx = np.r_[slice(*d.indices(s)) if isinstance(d, slice) else d]
        input_indices.append(idx)
        if not isinstance(d, Integral):
            value_indices.append(np.arange(len(idx)))
    return input_indices, value_indices


def compute_output_of_indexing(tensor, tensor_index):
    """
    Compute the output information of `__{set,get}item__` on tensor for every chunk.
    """
    tensor_index = process_index(tensor.ndim, tensor_index)
    output_shape = calc_shape(tensor.shape, tensor_index)

    index_tensor_op = TensorIndex(
        dtype=tensor.dtype, sparse=False, indexes=list(tensor_index)
    )
    index_tensor = tile(index_tensor_op.new_tensor([tensor], shape=tuple(output_shape)))
    output_chunks = index_tensor.chunks

    nsplits_acc = [
        np.cumsum(
            (0,)
            + tuple(
                c.shape[i]
                for c in output_chunks
                if all(idx == 0 for j, idx in enumerate(c.index) if j != i)
            )
        )
        for i in range(len(output_chunks[0].shape))
    ]
    return output_shape, output_chunks, nsplits_acc


def setitem_on_chunk_to_records(nsplits_acc, output_chunk, value, ts, is_scalar):
    """
    Turns a `__setitem__` on chunk to a list of index-value records.

    Parameters
    ----------
    nsplits_acc:
        Accumulate nsplits arrays of the output tensor chunks.

    Returns
    -------
        A list of `(index, value, timestamp)`, where `index` is the in-chunk index.
    """
    input_indices, value_indices = indexing_to_chunk_indices(output_chunk)

    # normalize assigned value
    if is_scalar:
        chunk_value = value
    else:
        chunk_value_slice = tuple(
            slice(
                nsplits_acc[i][output_chunk.index[i]],
                nsplits_acc[i][output_chunk.index[i] + 1],
            )
            for i in range(len(output_chunk.index))
        )
        chunk_value = value[chunk_value_slice]

    records = []
    for chunk_idx, value_idx in zip(
        itertools.product(*input_indices), itertools.product(*value_indices)
    ):
        new_value = chunk_value if is_scalar else chunk_value[value_idx]
        index_in_chunk = np.ravel_multi_index(
            chunk_idx, output_chunk.op.inputs[0].shape
        )
        records.append((index_in_chunk, new_value, ts))
    return records


def setitem_to_records(tensor, tensor_index, value, timestamp):
    """
    Compute the records of `__setitem__` on tensor for every chunk.

    Returns
    -------
        dict, a dict of chunk index to records in that chunk.
    """
    output_shape, output_chunks, nsplits_acc = compute_output_of_indexing(
        tensor, tensor_index
    )

    is_scalar = (
        np.isscalar(value)
        or isinstance(value, tuple)
        and tensor.dtype.fields is not None
    )
    if not is_scalar:
        value = np.broadcast_to(value, output_shape).astype(tensor.dtype)

    records = dict()
    for output_chunk in output_chunks:
        records_in_chunk = setitem_on_chunk_to_records(
            nsplits_acc, output_chunk, value, timestamp, is_scalar=is_scalar
        )
        records[output_chunk.op.inputs[0].index] = records_in_chunk
    return records


def getitem_on_chunk_to_records(nsplits_acc, output_chunk):
    """
    Turns a `__getitem__` on chunk to a list of index-value records.

    Parameters
    ----------
    nsplits_acc:
        Accumulate nsplits arrays of the output tensor chunks.

    Returns
    -------
        records: A list of `(index, value_index)`, where `index` is the in-chunk index, and
        `value_index` is the index in the final result block.
        chunk_value_shape: shape of result of this chunk.
        chunk_value_slice: index of result of this chunk in the whole result tensor.
    """
    input_indices, value_indices = indexing_to_chunk_indices(output_chunk)

    chunk_value_slice = tuple(
        slice(
            nsplits_acc[i][output_chunk.index[i]],
            nsplits_acc[i][output_chunk.index[i] + 1],
        )
        for i in range(len(output_chunk.index))
    )

    records = []
    for chunk_idx, value_idx in zip(
        itertools.product(*input_indices), itertools.product(*value_indices)
    ):
        index_in_chunk = np.ravel_multi_index(
            chunk_idx, output_chunk.op.inputs[0].shape
        )
        records.append((index_in_chunk, value_idx))
    return records, output_chunk.shape, chunk_value_slice


def getitem_to_records(tensor, tensor_index):
    """
    Compute the records of `__getitem__` on tensor for every chunk.

    Returns
    -------
        records and output_chunk dict, records is a dict of chunk index to records
        in that chunk.
    """
    output_shape, output_chunks, nsplits_acc = compute_output_of_indexing(
        tensor, tensor_index
    )

    records = dict()
    for output_chunk in output_chunks:
        records_in_chunk = getitem_on_chunk_to_records(nsplits_acc, output_chunk)
        records[output_chunk.op.inputs[0].index] = records_in_chunk
    return records, output_shape


def normalize_timestamp(timestamp=None):
    if timestamp is None:
        timestamp = np.datetime64(datetime.now())
    if isinstance(timestamp, datetime):
        timestamp = np.datetime64(timestamp)
    return timestamp


def normalize_name(name: Optional[str] = None):
    if not name:
        return str(uuid.uuid4())
    return name



from cython cimport floating
cimport numpy as np


cdef floating _euclidean_dense_dense(floating*, floating*, int, bint) nogil

cdef floating _euclidean_sparse_dense(floating[::1], int[::1], floating[::1],
                                      floating, bint) nogil

cdef void _average_centers(floating[:, ::1], floating[::1])

cdef void _center_shift(floating[:, ::1], floating[:, ::1], floating[::1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, StringField
from ...utils import has_unknown_shape
from ..utils import unify_chunks, broadcast_shape
from ..operands import TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from .broadcast_to import broadcast_to
from ..array_utils import as_same_device, device


class TensorCopyTo(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.COPYTO

    _src = KeyField("src")
    _dst = KeyField("dest")
    _casting = StringField("casting")
    _where = KeyField("where")

    def __init__(self, casting=None, **kw):
        super().__init__(_casting=casting, **kw)

    @property
    def src(self):
        return self._src

    @property
    def dst(self):
        return self._dst

    @property
    def casting(self):
        return self._casting

    @property
    def where(self):
        return self._where

    def check_inputs(self, inputs):
        if not 2 <= len(inputs) <= 3:
            raise ValueError("inputs' length must be 2 or 3")

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)

        self._src = self._inputs[0]
        self._dst = self._inputs[1]
        if len(self._inputs) > 2:
            self._where = self._inputs[2]

    @staticmethod
    def _extract_inputs(inputs):
        if len(inputs) == 2:
            (src, dst), where = inputs, None
        else:
            src, dst, where = inputs
            if where is True:
                where = None
            else:
                where = astensor(where)

        return src, dst, where

    def __call__(self, *inputs):
        from ..core import Tensor

        src, dst, where = self._extract_inputs(inputs)

        if not isinstance(dst, Tensor):
            raise TypeError("dst has to be a Tensor")

        self.dtype = dst.dtype
        self.gpu = dst.op.gpu
        self.sparse = dst.issparse()

        if not np.can_cast(src.dtype, dst.dtype, casting=self.casting):
            raise TypeError(
                f"Cannot cast array from {src.dtype!r} to {dst.dtype!r} "
                f"according to the rule {self.casting!s}"
            )

        try:
            broadcast_to(src, dst.shape)
        except ValueError:
            raise ValueError(
                "could not broadcast input array "
                f"from shape {src.shape!r} into shape {dst.shape!r}"
            )
        if where:
            try:
                broadcast_to(where, dst.shape)
            except ValueError:
                raise ValueError(
                    "could not broadcast where mask "
                    f"from shape {src.shape!r} into shape {dst.shape!r}"
                )

        inps = [src, dst]
        if where is not None:
            inps.append(where)
        ret = self.new_tensor(inps, dst.shape, order=dst.order)
        dst.data = ret.data

    @classmethod
    def tile(cls, op):
        if has_unknown_shape(*op.inputs):
            yield
        inputs = yield from unify_chunks(
            *[(input, list(range(input.ndim))[::-1]) for input in op.inputs]
        )
        output = op.outputs[0]

        chunk_shapes = [
            t.chunk_shape if hasattr(t, "chunk_shape") else t for t in inputs
        ]
        out_chunk_shape = broadcast_shape(*chunk_shapes)

        out_chunks = []
        nsplits = [[np.nan] * shape for shape in out_chunk_shape]
        get_index = lambda idx, t: tuple(
            0 if t.nsplits[i] == (1,) else ix for i, ix in enumerate(idx)
        )
        for out_idx in itertools.product(*(map(range, out_chunk_shape))):
            in_chunks = [
                t.cix[get_index(out_idx[-t.ndim :], t)] if t.ndim != 0 else t.chunks[0]
                for t in inputs
            ]
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk(
                    in_chunks,
                    shape=in_chunks[1].shape,
                    order=output.order,
                    index=out_idx,
                )
            )
            out_chunks.append(out_chunk)
            for i, idx, s in zip(itertools.count(0), out_idx, out_chunk.shape):
                nsplits[i][idx] = s

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            output.shape,
            order=output.order,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            dst = inputs[1].copy()
            src = inputs[0]
            where = inputs[2] if len(inputs) > 2 else True

            xp.copyto(dst, src, casting=op.casting, where=where)
            ctx[op.outputs[0].key] = dst


def copyto(dst, src, casting="same_kind", where=True):
    """
    Copies values from one array to another, broadcasting as necessary.

    Raises a TypeError if the `casting` rule is violated, and if
    `where` is provided, it selects which elements to copy.

    Parameters
    ----------
    dst : Tensor
        The tensor into which values are copied.
    src : array_like
        The tensor from which values are copied.
    casting : {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
        Controls what kind of data casting may occur when copying.

          * 'no' means the data types should not be cast at all.
          * 'equiv' means only byte-order changes are allowed.
          * 'safe' means only casts which can preserve values are allowed.
          * 'same_kind' means only safe casts or casts within a kind,
            like float64 to float32, are allowed.
          * 'unsafe' means any data conversions may be done.
    where : array_like of bool, optional
        A boolean tensor which is broadcasted to match the dimensions
        of `dst`, and selects elements to copy from `src` to `dst`
        wherever it contains the value True.
    """
    op = TensorCopyTo(casting=casting)
    return op(src, dst, where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any, Dict, Generator, List, Type, Tuple

from ..core import Serializer, Placeholder, buffered
from .field import Field, OneOfField


class SerializableMeta(type):
    def __new__(mcs, name: str, bases: Tuple[Type], properties: Dict):
        new_properties = dict()
        for base in bases:
            if hasattr(base, "_FIELDS"):
                new_properties.update(base._FIELDS)
        new_properties.update(properties)
        properties = new_properties

        property_to_fields = dict()
        # filter out all fields
        for k, v in properties.items():
            if not isinstance(v, Field):
                continue

            property_to_fields[k] = v
            properties[k] = v
            v._attr_name = k

        properties["_FIELDS"] = property_to_fields
        slots = set(properties.pop("__slots__", set()))
        if property_to_fields:
            slots.add("_FIELD_VALUES")
        properties["__slots__"] = tuple(slots)

        clz = type.__new__(mcs, name, bases, properties)
        return clz


class Serializable(metaclass=SerializableMeta):
    __slots__ = ()

    _FIELDS: Dict[str, Field]
    _FIELD_VALUES: Dict[str, Any]

    def __init__(self, *args, **kwargs):
        if args:  # pragma: no cover
            values = dict(zip(self.__slots__, args))
            values.update(kwargs)
            self._FIELD_VALUES = values
        else:
            self._FIELD_VALUES = kwargs

    def __repr__(self):
        values = ", ".join(
            [
                "{}={!r}".format(slot, getattr(self, slot, None))
                for slot in self.__slots__
            ]
        )
        return "{}({})".format(self.__class__.__name__, values)


class SerializableSerializer(Serializer):
    """
    Leverage DictSerializer to perform serde.
    """

    serializer_name = "serializable"

    @classmethod
    def _get_tag_to_values(cls, obj: Serializable):
        fields = obj._FIELDS
        attr_to_values = obj._FIELD_VALUES
        tag_to_values = dict()

        for field in fields.values():
            tag = field.tag
            attr_name = field.attr_name
            try:
                value = attr_to_values[attr_name]
            except KeyError:
                continue
            if field.on_serialize:
                value = field.on_serialize(value)
            tag_to_values[tag] = value

        return tag_to_values

    @buffered
    def serialize(self, obj: Serializable, context: Dict):
        tag_to_values = self._get_tag_to_values(obj)

        keys = [None] * len(tag_to_values)
        value_headers = [None] * len(tag_to_values)
        value_sizes = [0] * len(tag_to_values)
        value_buffers = []
        for idx, (key, val) in enumerate(tag_to_values.items()):
            keys[idx] = key
            value_headers[idx], val_buf = yield val
            value_sizes[idx] = len(val_buf)
            value_buffers.extend(val_buf)

        header = {
            "keys": keys,
            "value_headers": value_headers,
            "value_sizes": value_sizes,
            "class": type(obj),
        }
        return header, value_buffers

    def deserialize(
        self, header: Dict, buffers: List, context: Dict
    ) -> Generator[Any, Any, Serializable]:
        obj_class: Type[Serializable] = header.pop("class")

        tag_to_values = dict()
        pos = 0
        for key, value_header, value_size in zip(
            header["keys"], header["value_headers"], header["value_sizes"]
        ):
            tag_to_values[key] = (
                yield value_header,
                buffers[pos : pos + value_size],
            )  # noqa: E999
            pos += value_size

        attr_to_values = dict()
        for field in obj_class._FIELDS.values():
            try:
                value = attr_to_values[field.attr_name] = tag_to_values[field.tag]
            except KeyError:
                continue
            if not isinstance(field, OneOfField):
                if value is not None:
                    if field.on_deserialize:

                        def cb(v, field_):
                            attr_to_values[field_.attr_name] = field_.on_deserialize(v)

                    else:

                        def cb(v, field_):
                            attr_to_values[field_.attr_name] = v

                    if isinstance(value, Placeholder):
                        value.callbacks.append(partial(cb, field_=field))
                    else:
                        cb(value, field)

        obj = obj_class()
        obj._FIELD_VALUES = attr_to_values
        return obj


SerializableSerializer.register(Serializable)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="unary")
class TensorTrunc(TensorUnaryOp):
    _op_type_ = OperandDef.TRUNC
    _func_name = "trunc"


@infer_dtype(np.trunc)
def trunc(x, out=None, where=None, **kwargs):
    """
    Return the truncated value of the input, element-wise.

    The truncated value of the scalar `x` is the nearest integer `i` which
    is closer to zero than `x` is. In short, the fractional part of the
    signed number `x` is discarded.

    Parameters
    ----------
    x : array_like
        Input data.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor or scalar
        The truncated value of each element in `x`.

    See Also
    --------
    ceil, floor, rint

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([-1.7, -1.5, -0.2, 0.2, 1.5, 1.7, 2.0])
    >>> mt.trunc(a).execute()
    array([-1., -1., -0.,  0.,  1.,  1.,  2.])
    """
    op = TensorTrunc(**kwargs)
    return op(x, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import contextlib
import sys

from .file import AioFileObject, AioFilesystem
from .isolation import Isolation, new_isolation, get_isolation, stop_isolation
from .lru import alru_cache
from .parallelism import AioEvent


if sys.version_info[:2] < (3, 9):
    from ._threads import to_thread

    asyncio.to_thread = to_thread


if sys.version_info[:2] < (3, 7):
    # patch run and get_running_loop etc for python 3.6
    from ._runners import get_running_loop, run

    asyncio.run = run
    asyncio.get_running_loop = get_running_loop
    asyncio.create_task = asyncio.ensure_future

    # patch async generator
    from async_generator import asynccontextmanager

    contextlib.asynccontextmanager = asynccontextmanager



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import numpy as np


def make_import_error_func(package_name):
    def _func(*_, **__):  # pragma: no cover
        raise ImportError(
            f"Cannot import {package_name}, please reinstall " f"that package."
        )

    return _func


def pick_workers(workers, size):
    """
    Pick workers from a list.

    This method will try to pick workers as balanced as it can.

    1. If size <= len(workers), randomly pick workers from the list.
    2. If size > len(workers), just select all workers in a random order,
       then see the rest size, if it's still more than the workers size,
       return all workers in a random order, if not,
       randomly select workers from the list.

    :param workers: workers list
    :param size: number to pick from workers list
    :return: ndarray of selected workers whose length is `size`
    """
    result = np.empty(size, dtype=object)
    rest = size
    while rest > 0:
        start = size - rest
        to_pick_size = min(size - start, len(workers))
        result[start : start + to_pick_size] = np.random.permutation(workers)[
            :to_pick_size
        ]
        rest = rest - to_pick_size
    return result


def config_mod_getattr(mod_dict, globals_):
    def __getattr__(name):
        import importlib

        if name in mod_dict:
            mod_name, cls_name = mod_dict[name].rsplit(".", 1)
            mod = importlib.import_module(mod_name, globals_["__name__"])
            cls = globals_[name] = getattr(mod, cls_name)
            return cls
        else:  # pragma: no cover
            raise AttributeError(name)

    if sys.version_info[:2] < (3, 7):
        for _mod in mod_dict.keys():
            __getattr__(_mod)

    def __dir__():
        return sorted([n for n in globals_ if not n.startswith("_")] + list(mod_dict))

    globals_.update(
        {
            "__getattr__": __getattr__,
            "__dir__": __dir__,
            "__all__": list(__dir__()),
            "__warningregistry__": dict(),
        }
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, OutputType, recursive_tile
from ...serialization.serializables import (
    KeyField,
    SeriesField,
    DataTypeField,
    AnyField,
)
from ...tensor.datasource import tensor as astensor
from ...tensor.utils import unify_chunks
from ...utils import has_unknown_shape
from ..core import INDEX_TYPE, SERIES_TYPE, SERIES_CHUNK_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


class DataFrameFromTensor(DataFrameOperand, DataFrameOperandMixin):
    """
    Represents data from mars tensor
    """

    _op_type_ = OperandDef.DATAFRAME_FROM_TENSOR

    _input = AnyField("input")
    _index = KeyField("index")
    _dtypes = SeriesField("dtypes")

    def __init__(
        self, input_=None, index=None, dtypes=None, gpu=None, sparse=None, **kw
    ):
        super().__init__(
            _input=input_,
            _index=index,
            _dtypes=dtypes,
            gpu=gpu,
            sparse=sparse,
            _output_types=[OutputType.dataframe],
            **kw,
        )

    @property
    def dtypes(self):
        return self._dtypes

    @property
    def input(self):
        return self._input

    @property
    def index(self):
        return self._index

    def _tokenize_output(self, output_idx: int, **kw):
        # make sure all necessary arguments tokenized
        return None

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        if self._input is not None:
            if not isinstance(self._input, dict):
                self._input = next(inputs_iter)
            else:
                # check each value for input
                new_input = OrderedDict()
                for k, v in self._input.items():
                    if isinstance(v, ENTITY_TYPE):
                        new_input[k] = next(inputs_iter)
                    else:
                        new_input[k] = v
                self._input = new_input

        if self._index is not None:
            self._index = next(inputs_iter)

    def __call__(self, input_tensor, index, columns):
        if isinstance(input_tensor, dict):
            return self._call_input_1d_tileables(input_tensor, index, columns)
        elif input_tensor is not None:
            return self._call_input_tensor(input_tensor, index, columns)
        else:
            return self._call_tensor_none(index, columns)

    def _process_index(self, index, inputs):
        if not isinstance(index, pd.Index):
            if isinstance(index, INDEX_TYPE):
                self._index = index
                index_value = index.index_value
                inputs.append(index)
            elif isinstance(index, ENTITY_TYPE):
                self._index = index
                index = astensor(index)
                if index.ndim != 1:
                    raise ValueError(f"index should be 1-d, got {index.ndim}-d")
                index_value = parse_index(
                    pd.Index([], dtype=index.dtype), index, type(self).__name__
                )
                inputs.append(index)
            else:
                index = pd.Index(index)
                index_value = parse_index(index, store_data=True)
        else:
            index_value = parse_index(index, store_data=True)
        return index_value

    def _call_input_1d_tileables(self, input_1d_tileables, index, columns):
        tileables = []
        shape = None
        for tileable in input_1d_tileables.values():
            tileable_shape = astensor(tileable).shape
            if len(tileable_shape) > 0:
                if shape is None:
                    shape = tileable_shape
                elif shape != tileable_shape:
                    raise ValueError("input 1-d tensors should have same shape")

            if isinstance(tileable, ENTITY_TYPE):
                tileables.append(tileable)

        if index is not None:
            tileable_size = tileables[0].shape[0]
            if hasattr(index, "shape"):
                index_size = index.shape[0]
            else:
                index_size = len(index)
            if (
                not pd.isna(tileable_size)
                and not pd.isna(index_size)
                and tileable_size != index_size
            ):
                raise ValueError(
                    f"index {index} should have the same shape "
                    f"with tensor: {tileable_size}"
                )
            index_value = self._process_index(index, tileables)
        else:
            index_value = parse_index(pd.RangeIndex(0, tileables[0].shape[0]))

        if columns is not None:
            if len(input_1d_tileables) != len(columns):
                raise ValueError(
                    f"columns {columns} should have size {len(input_1d_tileables)}"
                )
            if not isinstance(columns, pd.Index):
                if isinstance(columns, ENTITY_TYPE):
                    raise NotImplementedError("The columns value cannot be a tileable")
                columns = pd.Index(columns)
            columns_value = parse_index(columns, store_data=True)
        else:
            columns_value = parse_index(
                pd.RangeIndex(0, len(input_1d_tileables)), store_data=True
            )

        shape = (shape[0], len(input_1d_tileables))
        return self.new_dataframe(
            tileables,
            shape,
            dtypes=self.dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    def _call_input_tensor(self, input_tensor, index, columns):
        if input_tensor.ndim not in {1, 2}:
            raise ValueError("Must pass 1-d or 2-d input")
        inputs = [input_tensor]

        if index is not None:
            if input_tensor.shape[0] != len(index):
                raise ValueError(
                    f"index {index} should have the same shape with tensor: {input_tensor.shape[0]}"
                )
            index_value = self._process_index(index, inputs)
        elif isinstance(input_tensor, SERIES_TYPE):
            index_value = input_tensor.index_value
        else:
            stop = input_tensor.shape[0]
            stop = -1 if np.isnan(stop) else stop
            index_value = parse_index(pd.RangeIndex(start=0, stop=stop))

        if columns is not None:
            if not (
                input_tensor.ndim == 1
                and len(columns) == 1
                or input_tensor.shape[1] == len(columns)
            ):
                raise ValueError(
                    f"columns {columns} should have the same shape with tensor: {input_tensor.shape[1]}"
                )
            if not isinstance(columns, pd.Index):
                if isinstance(columns, ENTITY_TYPE):
                    raise NotImplementedError("The columns value cannot be a tileable")
                columns = pd.Index(columns)
            columns_value = parse_index(columns, store_data=True)
        else:
            if input_tensor.ndim == 1:
                # convert to 1-d DataFrame
                columns_value = parse_index(
                    pd.RangeIndex(start=0, stop=1), store_data=True
                )
            else:
                columns_value = parse_index(
                    pd.RangeIndex(start=0, stop=input_tensor.shape[1]), store_data=True
                )

        if input_tensor.ndim == 1:
            shape = (input_tensor.shape[0], 1)
        else:
            shape = input_tensor.shape

        return self.new_dataframe(
            inputs,
            shape,
            dtypes=self.dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    def _call_tensor_none(self, index, columns):
        inputs = []
        shape = []
        if index is not None:
            index_value = self._process_index(index, inputs)
            shape.append(index.shape[0])
        else:
            index_value = parse_index(pd.Index([], dtype=object))
            shape.append(0)

        if columns is not None:
            if not isinstance(columns, pd.Index):
                if isinstance(columns, ENTITY_TYPE):
                    raise NotImplementedError("The columns value cannot be a tileable")
                columns = pd.Index(columns)
            columns_value = parse_index(columns, store_data=True)
            shape.append(columns.shape[0])
        else:
            columns_value = parse_index(pd.Index([], dtype=object), store_data=True)
            shape.append(0)

        return self.new_dataframe(
            inputs,
            shape=tuple(shape),
            dtypes=self.dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    @classmethod
    def tile(cls, op):
        if isinstance(op.input, dict):
            return (yield from cls._tile_input_1d_tileables(op))
        elif op.input is not None:
            return (yield from cls._tile_input_tensor(op))
        else:
            return cls._tile_tensor_none(op)

    @classmethod
    def _tile_input_1d_tileables(cls, op):
        # make sure all tensor have known chunk shapes
        if has_unknown_shape(*op.inputs):
            yield

        out_df = op.outputs[0]
        in_tensors = op.inputs
        in_tensors = yield from unify_chunks(*in_tensors)
        nsplit = in_tensors[0].nsplits[0]

        cum_sizes = [0] + np.cumsum(nsplit).tolist()
        out_chunks = []
        for i in range(in_tensors[0].chunk_shape[0]):
            chunk_op = op.copy().reset_key()
            new_input = OrderedDict()
            for k, v in op.input.items():
                if not isinstance(v, ENTITY_TYPE):
                    try:
                        new_input[k] = v[cum_sizes[i] : cum_sizes[i + 1]]
                    except TypeError:
                        # scalar
                        new_input[k] = v
                else:
                    # do not need to do slice,
                    # will be done in set_inputs
                    new_input[k] = v
            chunk_op._input = new_input
            columns_value = out_df.columns_value
            dtypes = out_df.dtypes
            chunk_index = (i, 0)
            if isinstance(op.index, INDEX_TYPE):
                index_value = in_tensors[-1].chunks[i].index_value
            elif out_df.index_value.has_value():
                pd_index = out_df.index_value.to_pandas()[
                    cum_sizes[i] : cum_sizes[i + 1]
                ]
                index_value = parse_index(pd_index, store_data=True)
            else:
                assert op.index is not None
                index_chunk = in_tensors[-1].cix[
                    i,
                ]
                index_value = parse_index(
                    pd.Index([], dtype=index_chunk.dtype),
                    index_chunk,
                    type(chunk_op).__name__,
                )
            shape = (nsplit[i], len(out_df.dtypes))
            out_chunk = chunk_op.new_chunk(
                [t.cix[(i,)] for t in in_tensors],
                shape=shape,
                index=chunk_index,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )
            out_chunks.append(out_chunk)

        nsplits = (nsplit, (len(out_df.dtypes),))
        new_op = op.copy()
        return new_op.new_dataframes(
            out_df.inputs,
            out_df.shape,
            dtypes=out_df.dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_input_tensor(cls, op):
        out_df = op.outputs[0]
        in_tensor = op.input
        out_chunks = []
        if out_df.index_value.has_value() and has_unknown_shape(in_tensor):
            yield

        nsplits = in_tensor.nsplits

        if op.index is not None:
            # rechunk index if it's a tensor
            if has_unknown_shape(*op.inputs):
                yield
            index_tensor = yield from recursive_tile(op.index.rechunk([nsplits[0]]))
        else:
            index_tensor = None

        # build column attrs for all chunks
        if in_tensor.ndim == 1:
            columns_value_list = [
                parse_index(out_df.columns_value.to_pandas()[0:1], store_data=True)
            ]
            dtypes_list = [out_df.dtypes]
        else:
            columns_value_list = [None] * in_tensor.chunk_shape[1]
            dtypes_list = [None] * in_tensor.chunk_shape[1]
            cum_cols = np.cumsum((0,) + in_tensor.nsplits[1])
            for idx in range(in_tensor.chunk_shape[1]):
                dtypes_list[idx] = out_df.dtypes[cum_cols[idx] : cum_cols[idx + 1]]
                pd_columns = out_df.columns_value.to_pandas()
                chunk_pd_columns = pd_columns[cum_cols[idx] : cum_cols[idx + 1]]
                columns_value_list[idx] = parse_index(chunk_pd_columns, store_data=True)

        cum_size = [np.cumsum(s) for s in nsplits]
        index_value_list = [None] * in_tensor.chunk_shape[0]
        for in_chunk in in_tensor.chunks:
            out_op = op.copy().reset_key()
            chunk_inputs = [in_chunk]
            if in_chunk.ndim == 1:
                (i,) = in_chunk.index
                column_stop = 1
                chunk_index = (in_chunk.index[0], 0)
                dtypes = dtypes_list[0]
                columns_value = columns_value_list[0]
                chunk_shape = (in_chunk.shape[0], 1)
            else:
                i, j = in_chunk.index
                column_stop = cum_size[1][j]
                chunk_index = in_chunk.index
                dtypes = dtypes_list[in_chunk.index[1]]
                columns_value = columns_value_list[in_chunk.index[1]]
                chunk_shape = in_chunk.shape

            index_stop = cum_size[0][i]
            if isinstance(op.index, INDEX_TYPE):
                index_chunk = index_tensor.chunks[i]
                index_value = index_chunk.index_value
                chunk_inputs.append(index_chunk)
            elif isinstance(in_chunk, SERIES_CHUNK_TYPE):
                index_value = in_chunk.index_value
            elif index_value_list[in_chunk.index[0]] is not None:
                # value already cached
                index_value = index_value_list[in_chunk.index[0]]
                if index_tensor is not None:
                    index_chunk = index_tensor.cix[
                        in_chunk.index[0],
                    ]
                    chunk_inputs.append(index_chunk)
            else:
                if out_df.index_value.has_value():
                    pd_index = out_df.index_value.to_pandas()
                    chunk_pd_index = pd_index[
                        index_stop - in_chunk.shape[0] : index_stop
                    ]
                    index_value = parse_index(chunk_pd_index, store_data=True)
                elif op.index is None:
                    # input tensor has unknown shape
                    index_value = parse_index(
                        pd.RangeIndex(-1), in_tensor, in_chunk.index[0]
                    )
                else:
                    index_chunk = index_tensor.cix[
                        in_chunk.index[0],
                    ]
                    chunk_inputs.append(index_chunk)
                    index_value = parse_index(
                        pd.Index([], dtype=index_tensor.dtype),
                        index_chunk,
                        type(out_op).__name__,
                    )
                index_value_list[in_chunk.index[0]] = index_value

            out_op.extra_params["index_stop"] = index_stop
            out_op.extra_params["column_stop"] = column_stop
            out_chunk = out_op.new_chunk(
                chunk_inputs,
                shape=chunk_shape,
                index=chunk_index,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )
            out_chunks.append(out_chunk)

        if in_tensor.ndim == 1:
            nsplits = in_tensor.nsplits + ((1,),)
        else:
            nsplits = in_tensor.nsplits

        new_op = op.copy()
        return new_op.new_dataframes(
            out_df.inputs,
            out_df.shape,
            dtypes=out_df.dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_tensor_none(cls, op: "DataFrameFromTensor"):
        out_df = op.outputs[0]

        out_chunks = []
        assert isinstance(op.index, INDEX_TYPE)
        # tile as index
        for index_chunk in op.index.chunks:
            index_value = index_chunk.index_value

            chunk_shape = (index_chunk.shape[0], out_df.shape[1])
            chunk_index = (index_chunk.index[0], 0)

            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk(
                [index_chunk],
                shape=chunk_shape,
                index=chunk_index,
                index_value=index_value,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_df.params.copy()
        params["nsplits"] = (op.index.nsplits[0], (out_df.shape[1],))
        params["chunks"] = out_chunks
        return new_op.new_dataframes(out_df.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]

        if isinstance(op.input, dict):
            d = OrderedDict()
            for k, v in op.input.items():
                if hasattr(v, "key"):
                    d[k] = ctx[v.key]
                else:
                    d[k] = v
            if op.index is not None:
                index_data = ctx[op.index.key]
            else:
                index_data = chunk.index_value.to_pandas()
            ctx[chunk.key] = pd.DataFrame(
                d, index=index_data, columns=chunk.columns_value.to_pandas()
            )
        elif op.input is not None:
            tensor_data = ctx[op.inputs[0].key]
            if isinstance(tensor_data, pd.Series):
                ctx[chunk.key] = tensor_data.to_frame(name=chunk.dtypes.index[0])
            else:
                if op.index is not None:
                    # index is a tensor
                    index_data = ctx[op.inputs[1].key]
                else:
                    index_data = chunk.index_value.to_pandas()
                    if isinstance(index_data, pd.RangeIndex) and len(index_data) == 0:
                        index_data = None
                ctx[chunk.key] = pd.DataFrame(
                    tensor_data,
                    index=index_data,
                    columns=chunk.columns_value.to_pandas(),
                )
        else:
            index_data = ctx[op.index.key]
            ctx[chunk.key] = pd.DataFrame(
                index=index_data, columns=chunk.columns_value.to_pandas()
            )


def dataframe_from_tensor(tensor, index=None, columns=None, gpu=None, sparse=False):
    if tensor is not None:
        if tensor.ndim > 2 or tensor.ndim <= 0:
            raise TypeError(
                f"Not support create DataFrame from {tensor.ndim} dims tensor"
            )
        try:
            col_num = tensor.shape[1]
        except IndexError:
            col_num = 1
        gpu = tensor.op.gpu if gpu is None else gpu
        dtypes = pd.Series([tensor.dtype] * col_num, index=columns)
    else:
        gpu = None
        if columns is not None:
            dtypes = pd.Series([], index=columns)
        else:
            dtypes = pd.Series([], index=pd.Index([], dtype=object))
    op = DataFrameFromTensor(input_=tensor, dtypes=dtypes, gpu=gpu, sparse=sparse)
    return op(tensor, index, columns)


def dataframe_from_1d_tileables(d, index=None, columns=None, gpu=None, sparse=False):
    if columns is not None:
        tileables = [d.get(c) for c in columns]
    else:
        columns = list(d.keys())
        tileables = list(d.values())

    gpu = (
        next((t.op.gpu for t in tileables if hasattr(t, "op")), False)
        if gpu is None
        else gpu
    )
    dtypes = pd.Series(
        [t.dtype if hasattr(t, "dtype") else pd.Series(t).dtype for t in tileables],
        index=columns,
    )
    op = DataFrameFromTensor(input_=d, dtypes=dtypes, gpu=gpu, sparse=sparse)
    return op(d, index, columns)


class SeriesFromTensor(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.SERIES_FROM_TENSOR

    _input = KeyField("input")
    _index = KeyField("index")
    _dtype = DataTypeField("dtype")

    def __init__(
        self, input_=None, index=None, dtype=None, gpu=None, sparse=None, **kw
    ):
        super().__init__(
            _input=input_,
            _index=index,
            _dtype=dtype,
            gpu=gpu,
            sparse=sparse,
            _output_types=[OutputType.series],
            **kw,
        )

    @property
    def input(self):
        return self._input

    @property
    def index(self):
        return self._index

    @property
    def dtype(self):
        return self._dtype

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self._input is not None:
            self._input = self._inputs[0]
        if self._index is not None:
            self._index = self._inputs[-1]

    @classmethod
    def tile(cls, op):
        if op.index is None:
            # check all inputs to make sure no unknown chunk shape
            if has_unknown_shape(*op.inputs):
                yield

        if op.input is None:
            return cls._tile_tensor_none(op)

        out_series = op.outputs[0]
        in_tensor = op.inputs[0]
        nsplits = in_tensor.nsplits

        if op.index is not None:
            index_tensor = yield from recursive_tile(op.index.rechunk([nsplits[0]]))
        else:
            index_tensor = None

        index_start = 0
        out_chunks = []
        series_index = out_series.index_value.to_pandas()
        for in_chunk in in_tensor.chunks:
            new_op = op.copy().reset_key()
            new_op.extra_params["index_start"] = index_start
            chunk_inputs = [in_chunk]
            if index_tensor is not None:
                index_chunk = index_tensor.cix[in_chunk.index]
                chunk_inputs.append(index_chunk)
                if isinstance(op.index, INDEX_TYPE):
                    index_value = index_chunk.index_value
                else:
                    index_value = parse_index(
                        pd.Index([], dtype=in_chunk.dtype),
                        index_chunk,
                        type(new_op).__name__,
                    )
            else:
                chunk_pd_index = series_index[
                    index_start : index_start + in_chunk.shape[0]
                ]
                index_value = parse_index(chunk_pd_index, store_data=True)
            index_start += in_chunk.shape[0]
            out_chunk = new_op.new_chunk(
                chunk_inputs,
                shape=in_chunk.shape,
                index=in_chunk.index,
                index_value=index_value,
                name=out_series.name,
                dtype=out_series.dtype,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs,
            shape=out_series.shape,
            dtype=out_series.dtype,
            index_value=out_series.index_value,
            name=out_series.name,
            chunks=out_chunks,
            nsplits=in_tensor.nsplits,
        )

    @classmethod
    def _tile_tensor_none(cls, op: "SeriesFromTensor"):
        out_series = op.outputs[0]

        out_chunks = []
        assert isinstance(op.index, INDEX_TYPE)
        # tile as index
        for index_chunk in op.index.chunks:
            index_value = index_chunk.index_value

            chunk_shape = (index_chunk.shape[0],)
            chunk_index = (index_chunk.index[0],)

            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk(
                [index_chunk],
                shape=chunk_shape,
                index=chunk_index,
                index_value=index_value,
                dtype=out_series.dtype,
                name=out_series.name,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_series.params.copy()
        params["nsplits"] = (op.index.nsplits[0],)
        params["chunks"] = out_chunks
        return new_op.new_tileables(out_series.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        if op.input is not None:
            tensor_data = ctx[op.input.key]
        else:
            tensor_data = None

        if op.index is not None:
            index_data = ctx[op.index.key]
        else:
            index_data = chunk.index_value.to_pandas()
            if (
                tensor_data is not None
                and isinstance(index_data, pd.RangeIndex)
                and len(index_data) == 0
            ):
                # index not specified
                index_data = None

        ctx[chunk.key] = pd.Series(
            tensor_data, index=index_data, name=chunk.name, dtype=chunk.dtype
        )

    def __call__(self, input_tensor, index, name):
        inputs = [input_tensor] if input_tensor is not None else []
        if index is not None:
            if not isinstance(index, pd.Index):
                if isinstance(index, INDEX_TYPE):
                    self._index = index
                    index_value = index.index_value
                    inputs.append(index)
                elif isinstance(index, ENTITY_TYPE):
                    self._index = index
                    index = astensor(index)
                    if index.ndim != 1:
                        raise ValueError(f"index should be 1-d, got {index.ndim}-d")
                    index_value = parse_index(
                        pd.Index([], dtype=index.dtype), index, type(self).__name__
                    )
                    inputs.append(index)
                else:
                    index = pd.Index(index)
                    index_value = parse_index(index, store_data=True)
            else:
                index_value = parse_index(index, store_data=True)
        elif input_tensor is not None:
            if pd.isna(input_tensor.shape[0]):
                pd_index_value = pd.RangeIndex(-1)
            else:
                pd_index_value = pd.RangeIndex(start=0, stop=input_tensor.shape[0])
            index_value = parse_index(pd_index_value)
        else:
            index_value = parse_index(pd.Index([], dtype=object))

        if input_tensor is not None:
            shape = input_tensor.shape
        elif index is not None:
            shape = index.shape
        else:
            shape = (0,)

        return self.new_series(
            inputs, shape=shape, dtype=self.dtype, index_value=index_value, name=name
        )


def series_from_tensor(
    tensor, index=None, name=None, dtype=None, gpu=None, sparse=False
):
    if tensor is not None:
        if tensor.ndim > 1 or tensor.ndim <= 0:
            raise TypeError(f"Not support create Series from {tensor.ndim} dims tensor")
        gpu = tensor.op.gpu if gpu is None else gpu
        dtype = dtype or tensor.dtype
    else:
        gpu = None
        dtype = dtype or np.dtype(float)
    op = SeriesFromTensor(input_=tensor, dtype=dtype, gpu=gpu, sparse=sparse)
    return op(tensor, index, name)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import platform
import socket
import sys
from typing import Dict, List

import numpy as np
import pandas as pd

try:
    import scipy
except ImportError:  # pragma: no cover
    scipy = None

from ... import resource as mars_resource
from ...config import options
from ...utils import git_info, lazy_import
from ...storage import StorageLevel
from .core import WorkerSlotInfo, QuotaInfo, DiskInfo, StorageInfo

cp = lazy_import("cupy", globals=globals(), rename="cp")
cudf = lazy_import("cudf", globals=globals())

logger = logging.getLogger(__name__)

_is_initial = True


def gather_node_env():
    from ...lib.mkl_interface import mkl_get_version
    from ...lib.nvutils import NVError
    from ..._version import __version__ as mars_version

    global _is_initial
    if _is_initial:
        _is_initial = False
        mars_resource.cpu_percent()

    mem_stats = mars_resource.virtual_memory()

    node_info = {
        "command_line": sys.argv,
        "platform": platform.platform(),
        "host_name": socket.gethostname(),
        "python_version": sys.version,
        "mars_version": mars_version,
        "cpu_total": mars_resource.cpu_count(),
        "memory_total": mem_stats.total,
        "options": options.to_dict(),
    }

    if "MARS_K8S_POD_NAME" in os.environ:
        node_info["k8s_pod_name"] = os.environ["MARS_K8S_POD_NAME"]
    if "CONTAINER_ID" in os.environ:
        node_info["yarn_container_id"] = os.environ["CONTAINER_ID"]

    try:
        cuda_info = mars_resource.cuda_info()
    except NVError:  # pragma: no cover
        logger.exception("NVError encountered, cannot gather CUDA devices.")
        cuda_info = None

    if cuda_info:
        node_info["cuda_info"] = {
            "driver": cuda_info.driver_version,
            "cuda": cuda_info.cuda_version,
            "products": list(cuda_info.products),
        }

    package_vers = {
        "numpy": np.__version__,
        "pandas": pd.__version__,
    }
    if hasattr(np, "__mkl_version__") and mkl_get_version:
        mkl_version = mkl_get_version()
        package_vers[
            "mkl"
        ] = f"{mkl_version.major}.{mkl_version.minor}.{mkl_version.update}"

    if scipy is not None:
        package_vers["scipy"] = scipy.__version__
    if cp is not None:
        package_vers["cupy"] = cp.__version__
    if cudf is not None:
        package_vers["cudf"] = cudf.__version__

    node_info["package_versions"] = package_vers

    git = git_info()
    if git:
        node_info["git_info"] = {
            "hash": git.commit_hash,
            "ref": git.commit_ref,
        }

    bands = node_info["bands"] = dict()

    cpu_band = {
        "resources": {
            "cpu": mars_resource.cpu_count(),
            "memory": mars_resource.virtual_memory().total,
        }
    }
    # todo numa can be supported by adding more bands
    bands["numa-0"] = cpu_band

    for idx, gpu_card_stat in enumerate(
        mars_resource.cuda_card_stats()
    ):  # pragma: no cover
        bands[f"gpu-{idx}"] = {
            "resources": {
                "gpu": 1,
                "memory": gpu_card_stat.fb_mem_info.total,
            }
        }
    return node_info


def gather_node_resource(band_to_slots=None, use_gpu=True):
    # todo numa can be supported by adding more bands
    res = dict()
    mem_info = mars_resource.virtual_memory()
    num_cpu = (
        mars_resource.cpu_count()
        if band_to_slots is None
        else band_to_slots.get("numa-0", None)
    )
    if num_cpu:  # pragma: no branch
        res["numa-0"] = {
            "cpu_avail": mars_resource.cpu_count()
            - mars_resource.cpu_percent() / 100.0,
            "cpu_total": num_cpu,
            "memory_avail": mem_info.available,
            "memory_total": mem_info.total,
        }

    if use_gpu:
        for idx, gpu_card_stat in enumerate(
            mars_resource.cuda_card_stats()
        ):  # pragma: no cover
            num_gpu = 1 if band_to_slots is None else band_to_slots.get(f"gpu-{idx}")
            if not num_gpu:
                continue
            res[f"gpu-{idx}"] = {
                "gpu_avail": 1 - gpu_card_stat.gpu_usage,
                "gpu_total": num_gpu,
                "memory_avail": gpu_card_stat.fb_mem_info.available,
                "memory_total": gpu_card_stat.fb_mem_info.total,
            }
    return res


def gather_node_details(
    band_slot_infos: Dict[str, List[WorkerSlotInfo]] = None,
    band_quota_infos: Dict[str, QuotaInfo] = None,
    disk_infos: List[DiskInfo] = None,
    band_storage_infos: Dict[str, Dict[StorageLevel, StorageInfo]] = None,
):
    disk_io_usage = mars_resource.disk_io_usage()
    net_io_usage = mars_resource.net_io_usage()
    res = {
        "disk": dict(zip(("reads", "writes"), disk_io_usage))
        if disk_io_usage
        else dict(),
        "network": dict(zip(("receives", "sends"), net_io_usage))
        if net_io_usage
        else dict(),
        "iowait": mars_resource.iowait(),
    }

    if disk_infos:
        part_dict = dict()
        for info in disk_infos:
            part_dev = mars_resource.get_path_device(info.path)
            if part_dev in part_dict:
                continue

            disk_usage_result = mars_resource.disk_usage(info.path)
            io_usage_result = mars_resource.disk_io_usage(info.path)
            part_dict[part_dev] = disk_info = {
                "size_limit": info.limit_size,
                "size_used": disk_usage_result.used,
                "size_total": disk_usage_result.total,
            }
            if io_usage_result is not None:
                disk_info.update(
                    {
                        "reads": io_usage_result.reads if io_usage_result else None,
                        "writes": io_usage_result.writes if io_usage_result else None,
                    }
                )
            if not sys.platform.startswith("win"):
                in_usage_result = os.statvfs(info.path)
                disk_info.update(
                    {
                        "inode_used": in_usage_result.f_files
                        - in_usage_result.f_favail,
                        "inode_total": in_usage_result.f_files,
                    }
                )
        res["disk"]["partitions"] = part_dict

    band_slot_infos = band_slot_infos or dict()
    res["slot"] = {
        band: [
            {
                "slot_id": slot_info.slot_id,
                "session_id": slot_info.session_id,
                "subtask_id": slot_info.subtask_id,
                "processor_usage": slot_info.processor_usage,
            }
            for slot_info in slot_infos
        ]
        for band, slot_infos in band_slot_infos.items()
    }

    band_quota_infos = band_quota_infos or dict()
    res["quota"] = {
        band: {
            "quota_size": quota_info.quota_size,
            "allocated_size": quota_info.allocated_size,
            "hold_size": quota_info.hold_size,
        }
        for band, quota_info in band_quota_infos.items()
    }

    band_storage_infos = band_storage_infos or dict()
    res["storage"] = {
        band: {
            level.name.lower(): {
                "size_used": storage_info.used_size,
                "size_total": storage_info.total_size,
                "size_pinned": storage_info.pinned_size,
            }
            for level, storage_info in storage_infos.items()
        }
        for band, storage_infos in band_storage_infos.items()
    }
    return res



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess

import pytest

from mars.config import option_context
from mars.core.mode import is_kernel_mode, is_build_mode
from mars.lib.aio import stop_isolation
from mars.oscar.backends.router import Router
from mars.oscar.backends.ray.communication import RayServer
from mars.serialization.ray import register_ray_serializers, unregister_ray_serializers
from mars.utils import lazy_import

ray = lazy_import("ray")


@pytest.fixture
def ray_start_regular(request):
    param = getattr(request, "param", {})
    if not param.get("enable", True):
        yield
    else:
        register_ray_serializers()
        try:
            yield ray.init(num_cpus=20)
        finally:
            ray.shutdown()
            unregister_ray_serializers()
            Router.set_instance(None)
            RayServer.clear()
            if "COV_CORE_SOURCE" in os.environ:
                # Remove this when https://github.com/ray-project/ray/issues/16802 got fixed
                subprocess.check_call(["ray", "stop", "--force"])


@pytest.fixture
def ray_large_cluster(request):  # pragma: no cover
    param = getattr(request, "param", {})
    num_nodes = param.get("num_nodes", 3)
    num_cpus = param.get("num_cpus", 16)
    try:
        from ray.cluster_utils import Cluster
    except ModuleNotFoundError:
        from ray._private.cluster_utils import Cluster
    cluster = Cluster()
    remote_nodes = []
    for i in range(num_nodes):
        remote_nodes.append(
            cluster.add_node(num_cpus=num_cpus, memory=num_cpus * 2 * 1024**3)
        )
        if len(remote_nodes) == 1:
            ray.init(address=cluster.address)
    register_ray_serializers()
    try:
        yield
    finally:
        unregister_ray_serializers()
        Router.set_instance(None)
        RayServer.clear()
        ray.shutdown()
        cluster.shutdown()
        if "COV_CORE_SOURCE" in os.environ:
            # Remove this when https://github.com/ray-project/ray/issues/16802 got fixed
            subprocess.check_call(["ray", "stop", "--force"])


@pytest.fixture
def stop_ray(request):  # pragma: no cover
    yield
    if ray.is_initialized():
        ray.shutdown()


@pytest.fixture
async def ray_create_mars_cluster(request):
    from mars.deploy.oscar.ray import new_cluster, _load_config

    ray_config = _load_config()
    param = getattr(request, "param", {})
    supervisor_mem = param.get("supervisor_mem", 1 * 1024**3)
    worker_num = param.get("worker_num", 2)
    worker_cpu = param.get("worker_cpu", 2)
    worker_mem = param.get("worker_mem", 256 * 1024**2)
    ray_config.update(param.get("config", {}))
    client = await new_cluster(
        "test_cluster",
        supervisor_mem=supervisor_mem,
        worker_num=worker_num,
        worker_cpu=worker_cpu,
        worker_mem=worker_mem,
        config=ray_config,
    )
    async with client:
        yield client


@pytest.fixture(scope="module")
def _stop_isolation():
    yield
    stop_isolation()


@pytest.fixture(scope="module")
def _new_test_session(_stop_isolation):
    from .deploy.oscar.tests.session import new_test_session

    sess = new_test_session(
        address="test://127.0.0.1", init_local=True, default=True, timeout=300
    )
    with option_context({"show_progress": False}):
        try:
            yield sess
        finally:
            sess.stop_server(isolation=False)


@pytest.fixture(scope="module")
def _new_integrated_test_session(_stop_isolation):
    from .deploy.oscar.tests.session import new_test_session

    sess = new_test_session(
        address="127.0.0.1", init_local=True, n_worker=2, default=True, timeout=300
    )
    with option_context({"show_progress": False}):
        try:
            yield sess
        finally:
            sess.stop_server(isolation=False)


@pytest.fixture(scope="module")
def _new_gpu_test_session(_stop_isolation):  # pragma: no cover
    from .deploy.oscar.tests.session import new_test_session
    from .resource import cuda_count

    cuda_devices = list(range(min(cuda_count(), 2)))

    sess = new_test_session(
        address="127.0.0.1",
        init_local=True,
        n_worker=1,
        n_cpu=1,
        cuda_devices=cuda_devices,
        default=True,
        timeout=300,
    )
    with option_context({"show_progress": False}):
        try:
            yield sess
        finally:
            sess.stop_server(isolation=False)


@pytest.fixture
def setup(_new_test_session):
    _new_test_session.as_default()
    yield _new_test_session
    assert not (is_build_mode() or is_kernel_mode())


@pytest.fixture
def setup_cluster(_new_integrated_test_session):
    _new_integrated_test_session.as_default()
    yield _new_integrated_test_session


@pytest.fixture
def setup_gpu(_new_gpu_test_session):  # pragma: no cover
    _new_gpu_test_session.as_default()
    yield _new_test_session
